/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Ray.h"
#include "Transclip.h"
#include "ClipFlags.h"

#include "Bound.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
Bound::Bound()
 :	position_m(),
	dimension_m()
{
}


Bound::Bound
(
	const Vector3f& position,
	const Vector3f& dimension
)
{
	Bound::setPosDim( position, dimension );
}


Bound::~Bound()
{
}


Bound::Bound( const Bound& other )
{
	Bound::operator=( other );
}


Bound& Bound::operator=( const Bound& other )
{
	if( &other != this )
	{
		position_m  = other.position_m;
		dimension_m = other.dimension_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Bound::setPosDim
(
	const Vector3f& position,
	const Vector3f& dimension
)
{
	position_m  = position;
	dimension_m = dimension;
}


void Bound::setMinMax
(
	const Vector3f& min,
	const Vector3f& max
)
{
	setPosDim( min, max - min );
}




/// queries ----------------------------------------------------------------------------------------
bool Bound::getIntersection
(
	const Ray& ray
) const
{
	/// ch2 p65, 'an introduction to ray tracing' - glassner (AP 89)

	/// inverse-transform ray
	const Vector3f rayDirection( ray.getDirection().getVector() );
	Vector3f       rayOrigin( ray.getPosition() );
	rayOrigin -= position_m;


	float tNear = -FLOAT_MAX;
	float tFar  = +FLOAT_MAX;

	/// X planes
	if( rayDirection.getX() == 0.0f )
	{
		if( (rayOrigin.getX() < 0.0f) | (rayOrigin.getX() > dimension_m.getX()) )
		{
			return false;
		}
	}
	else
	{
		const float oneOverX = 1.0f / rayDirection.getX();
		float t1 = -rayOrigin.getX() * oneOverX;
		float t2 = (dimension_m.getX() - rayOrigin.getX()) * oneOverX;
		if( t1 > t2 )
		{
			const float temp = t1;
			t1 = t2;
			t2 = temp;
		}
		tNear = t1;
		tFar  = t2;
		if( tFar < 0 )
		{
			return false;
		}
	}

	/// Y planes
	if( rayDirection.getY() == 0.0f )
	{
		if( (rayOrigin.getY() < 0.0f) | (rayOrigin.getY() > dimension_m.getY()) )
		{
			return false;
		}
	}
	else
	{
		const float oneOverY = 1.0f / rayDirection.getY();
		float t1 = -rayOrigin.getY() * oneOverY;
		float t2 = (dimension_m.getY() - rayOrigin.getY()) * oneOverY;
		if( t1 > t2 )
		{
			const float temp = t1;
			t1 = t2;
			t2 = temp;
		}
		if( t1 > tNear )
		{
			tNear = t1;
		}
		if( t2 < tFar )
		{
			tFar = t2;
		}
		if( (tNear > tFar) | (tFar < 0) )
		{
			return false;
		}
	}

	/// Z planes
	if( rayDirection.getZ() == 0.0f )
	{
		if( (rayOrigin.getZ() < 0.0f) | (rayOrigin.getZ() > dimension_m.getZ()) )
		{
			return false;
		}
	}
	else
	{
		const float oneOverZ = 1.0f / rayDirection.getZ();
		float t1 = -rayOrigin.getZ() * oneOverZ;
		float t2 = (dimension_m.getZ() - rayOrigin.getZ()) * oneOverZ;
		if( t1 > t2 )
		{
			const float temp = t1;
			t1 = t2;
			t2 = temp;
		}
		if( t1 > tNear )
		{
			tNear = t1;
		}
		if( t2 < tFar )
		{
			tFar = t2;
		}
		if( (tNear > tFar) | (tFar < 0) )
		{
			return false;
		}
	}

	return true;
}


/*bool Bound::getIntersection
(
	const Ray& ray
) const
{
	/// schroeder, gdmag
	/// adjusted

	/// cohen-sutherland clipping codes
	static const dword CLIP_RIGHT  = 1 << 0;
	static const dword CLIP_LEFT   = 1 << 1;
	static const dword CLIP_TOP    = 1 << 2;
	static const dword CLIP_BOTTOM = 1 << 3;
	static const dword CLIP_BACK   = 1 << 4;
	static const dword CLIP_FRONT  = 1 << 5;


	const Vector3f rayDirection( ray.getDirection().getVector() );
	const Vector3f rayOrigin( ray.getPosition() );
	const Vector3f boundMin( position_m );
	const Vector3f boundMax( position_m + dimension_m );

	Vector3f intercept;


	dword outcode1 = 0;
	{
		const float x0 = boundMax.getX() - rayOrigin.getX();
		outcode1 |= (reinterpret_cast<const dword&>( x0 ) >> 31) & CLIP_RIGHT;
		const float x1 = rayOrigin.getX() - boundMin.getX();
		outcode1 |= (reinterpret_cast<const dword&>( x1 ) >> 31) & CLIP_LEFT;

		const float y0 = boundMax.getY() - rayOrigin.getY();
		outcode1 |= (reinterpret_cast<const dword&>( y0 ) >> 31) & CLIP_TOP;
		const float y1 = rayOrigin.getY() - boundMin.getY();
		outcode1 |= (reinterpret_cast<const dword&>( y1 ) >> 31) & CLIP_BOTTOM;

		const float z0 = boundMax.getZ() - rayOrigin.getZ();
		outcode1 |= (reinterpret_cast<const dword&>( z0 ) >> 31) & CLIP_BACK;
		const float z1 = rayOrigin.getZ() - boundMin.getZ();
		outcode1 |= (reinterpret_cast<const dword&>( z1 ) >> 31) & CLIP_FRONT;

		/// 3% slower
		//if( rayOrigin.getX() > boundMax.getX() )
		//{
		//	outcode1 |= CLIP_RIGHT;
		//}
		//else if( rayOrigin.getX() < boundMin.getX() )
		//{
		//	outcode1 |= CLIP_LEFT;
		//}
		//
		//if( rayOrigin.getY() > boundMax.getY() )
		//{
		//	outcode1 |= CLIP_TOP;
		//}
		//else if( rayOrigin.getY() < boundMin.getY() )
		//{
		//	outcode1 |= CLIP_BOTTOM;
		//}
		//
		//if( rayOrigin.getZ() > boundMax.getZ() )
		//{
		//	outcode1 |= CLIP_BACK;
		//}
		//else if( rayOrigin.getZ() < boundMin.getZ() )
		//{
		//	outcode1 |= CLIP_FRONT;
		//}
	}
	if( outcode1 == 0 )
	{
		/// origin inside bound
		intercept = rayOrigin;
		return true;
	}

	{
		dword outcode2 = 0;

		//const dword dirX = reinterpret_cast<const dword&>( rayDirection.getX() );
		//outcode2 |= (~(dirX >> 31) & CLIP_RIGHT) | ((dirX >> 31) & CLIP_LEFT);
		//outcode2 |= (((dirX | 0x7FFFFFFF) - 1) >> 31) & (CLIP_RIGHT | CLIP_LEFT);

		if( rayDirection.getX() >= 0.0f )
		{
			outcode2 |= CLIP_RIGHT;
		}
		if( rayDirection.getX() <= 0.0f )
		{
			outcode2 |= CLIP_LEFT;
		}

		if( rayDirection.getY() >= 0.0f )
		{
			outcode2 |= CLIP_TOP;
		}
		if( rayDirection.getY() <= 0.0f )
		{
			outcode2 |= CLIP_BOTTOM;
		}

		if( rayDirection.getZ() >= 0.0f )
		{
			outcode2 |= CLIP_BACK;
		}
		if( rayDirection.getZ() <= 0.0f )
		{
			outcode2 |= CLIP_FRONT;
		}

		if( (outcode1 & outcode2) != 0 )
		{
			/// ray not pointing toward bound
			return false;
		}
	}

	/// check intersections
	if( outcode1 & (CLIP_RIGHT | CLIP_LEFT) )
	{
		if( outcode1 & CLIP_RIGHT )
		{
			intercept.x_m = boundMax.getX();
		}
		else
		{
			intercept.x_m = boundMin.getX();
		}
		const float xd = intercept.getX() - rayOrigin.getX();
		const float oneOverDirectionX = 1.0f / rayDirection.getX();
		intercept.y_m = rayOrigin.getY() + xd * rayDirection.getY() * oneOverDirectionX;
		intercept.z_m = rayOrigin.getZ() + xd * rayDirection.getZ() * oneOverDirectionX;

		if( (intercept.getY() < boundMax.getY()) & (intercept.getY() >= boundMin.getY()) &
		    (intercept.getZ() < boundMax.getZ()) & (intercept.getZ() >= boundMin.getZ()) )
		{
			return true;
		}
	}

	if( outcode1 & (CLIP_TOP | CLIP_BOTTOM) )
	{
		if( outcode1 & CLIP_TOP )
		{
			intercept.y_m = boundMax.getY();
		}
		else
		{
			intercept.y_m = boundMin.getY();
		}
		const float yd = intercept.getY() - rayOrigin.getY();
		const float oneOverDirectionY = 1.0f / rayDirection.getY();
		intercept.x_m = rayOrigin.getX() + yd * rayDirection.getX() * oneOverDirectionY;
		intercept.z_m = rayOrigin.getZ() + yd * rayDirection.getZ() * oneOverDirectionY;

		if( (intercept.getX() < boundMax.getX()) & (intercept.getX() >= boundMin.getX()) &
		    (intercept.getZ() < boundMax.getZ()) & (intercept.getZ() >= boundMin.getZ()) )
		{
			return true;
		}
	}

	if( outcode1 & (CLIP_BACK | CLIP_FRONT) )
	{
		if( outcode1 & CLIP_BACK )
		{
			intercept.z_m = boundMax.getZ();
		}
		else
		{
			intercept.z_m = boundMin.getZ();
		}
		const float zd = intercept.getZ() - rayOrigin.getZ();
		const float oneOverDirectionZ = 1.0f / rayDirection.getZ();
		intercept.x_m = rayOrigin.getX() + zd * rayDirection.getX() * oneOverDirectionZ;
		intercept.y_m = rayOrigin.getY() + zd * rayDirection.getY() * oneOverDirectionZ;

		if( (intercept.getX() < boundMax.getX()) & (intercept.getX() >= boundMin.getX()) &
		    (intercept.getY() < boundMax.getY()) & (intercept.getY() >= boundMin.getY()) )
		{
			return true;
		}
	}


	/// no intersection
	return false;
}*/


/*bool Bound::getIntersection
(
	const Ray& ray
) const
{
	/// original code from:
	/// 'fast ray-box intersection' - woo (graphics gems, academic press, 1990)
	/// optimized code from:
	/// pierre terdiman, p.terdiman@codercorner.com, 2000 (~20-30% faster on a celeron 500)
	/// then adjusted to fit here.

	bool isInside = true;

	const Vector3f rayDirection( ray.getDirection().getVector() );
	const Vector3f rayOrigin( ray.getPosition() );
	const Vector3f boundMin( position_m );
	const Vector3f boundMax( position_m + dimension_m );

	float maxDistance[3] = { -1.0f, -1.0f, -1.0f };
	Vector3f hitCoords;

	/// find candidate planes
	{
		/// x
		if( rayOrigin.getX() < boundMin.getX() )
		{
			hitCoords.x_m = boundMin.getX();
			isInside      = false;

			if( rayDirection.getX() != 0.0f )
			{
				maxDistance[0] = (boundMin.getX() - rayOrigin.getX()) / rayDirection.getX();
			}
		}
		else
		if( rayOrigin.getX() > boundMax.getX() )
		{
			hitCoords.x_m = boundMax.getX();
			isInside      = false;

			if( rayDirection.getX() != 0.0f )
			{
				maxDistance[0] = (boundMax.getX() - rayOrigin.getX()) / rayDirection.getX();
			}
		}

		/// y
		if( rayOrigin.getY() < boundMin.getY() )
		{
			hitCoords.y_m = boundMin.getY();
			isInside      = false;

			if( rayDirection.getY() != 0.0f )
			{
				maxDistance[1] = (boundMin.getY() - rayOrigin.getY()) / rayDirection.getY();
			}
		}
		else
		if( rayOrigin.getY() > boundMax.getY() )
		{
			hitCoords.y_m = boundMax.getY();
			isInside      = false;

			if( rayDirection.getY() != 0.0f )
			{
				maxDistance[1] = (boundMax.getY() - rayOrigin.getY()) / rayDirection.getY();
			}
		}

		/// z
		if( rayOrigin.getZ() < boundMin.getZ() )
		{
			hitCoords.z_m = boundMin.getZ();
			isInside      = false;

			if( rayDirection.getZ() != 0.0f )
			{
				maxDistance[2] = (boundMin.getZ() - rayOrigin.getZ()) / rayDirection.getZ();
			}
		}
		else
		if( rayOrigin.getZ() > boundMax.getZ() )
		{
			hitCoords.z_m = boundMax.getZ();
			isInside      = false;

			if( rayDirection.getZ() != 0.0f )
			{
				maxDistance[2] = (boundMax.getZ() - rayOrigin.getZ()) / rayDirection.getZ();
			}
		}
	}

	/// ray origin inside bounding box
	if( isInside )
	{
		hitCoords = rayOrigin;
		return true;
	}

	/// get largest of the maxDistances for final choice of intersection
	dword whichPlane;// = 0;
	if( maxDistance[1] > maxDistance[0] )
	{
		whichPlane = (maxDistance[2] > maxDistance[1]) ? 2 : 1;
	}
	else
	{
		whichPlane = (maxDistance[2] > maxDistance[0]) ? 2 : 0;
	}

	/// check final candidate actually inside box
	if( maxDistance[whichPlane] < 0.0f )
	{
		return false;
	}

	{
		if( whichPlane != 0 )
		{
			hitCoords.x_m = rayOrigin.getX() + maxDistance[whichPlane] * rayDirection.getX();
			if( (hitCoords.getX() <  boundMin.getX()) |
			    (hitCoords.getX() >= boundMax.getX()) )
			{
				return false;
			}
		}

		if( whichPlane != 1 )
		{
			hitCoords.y_m = rayOrigin.getY() + maxDistance[whichPlane] * rayDirection.getY();
			if( (hitCoords.getY() <  boundMin.getY()) |
			    (hitCoords.getY() >= boundMax.getY()) )
			{
				return false;
			}
		}

		if( whichPlane != 2 )
		{
			hitCoords.z_m = rayOrigin.getZ() + maxDistance[whichPlane] * rayDirection.getZ();
			if( (hitCoords.getZ() <  boundMin.getZ()) |
			    (hitCoords.getZ() >= boundMax.getZ()) )
			{
				return false;
			}
		}
	}

	// ray hits
	return true;
}*/


bool Bound::getClipping
(
	const ClipFlags& enclosingClipFlags,
	const Transclip& transform,
	ClipFlags&       clipFlags
) const
{
	Vector3f boundCorners[8];
	getCorners( boundCorners );

	return transform.getClipping( enclosingClipFlags, boundCorners, clipFlags );
}


void Bound::getMinMax
(
	Vector3f& min,
	Vector3f& max
) const
{
	min = position_m;
	(max = position_m) += dimension_m;
}


void Bound::getCorners
(
	Vector3f corners[8]
) const
{
	corners[0]  = position_m;
	(corners[1] = position_m) += Vector3f( dimension_m.getX(), 0.0f,               0.0f );
	(corners[2] = position_m) += Vector3f( 0.0f,               dimension_m.getY(), 0.0f );
	(corners[3] = position_m) += Vector3f( dimension_m.getX(), dimension_m.getY(), 0.0f );
	(corners[4] = position_m) += Vector3f( 0.0f,               0.0f,               dimension_m.getZ() );
	(corners[5] = position_m) += Vector3f( dimension_m.getX(), 0.0f,               dimension_m.getZ() );
	(corners[6] = position_m) += Vector3f( 0.0f,               dimension_m.getY(), dimension_m.getZ() );
	(corners[7] = position_m) += dimension_m;
}
