/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Scene.h"
#include "Bound.h"
#include "ObjectRenderableId.h"
#include "Ray.h"


#include "OctreeVisitorTracingSubpart.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
OctreeVisitorTracingSubpart::OctreeVisitorTracingSubpart
(
	const Scene* pScene
)
 :	OctreeVisitorT<SubPart>(),
	pScene_m               ( pScene ),
	pLastSubpart_m       ( 0 ),
	subpartUntransLower_m(),
	subpartUntransUpper_m(),
	//subpartTransBound_m[](),
	subpartInvTransform_m()
//	pObjectId_m   (),
//	pRay_m        (),
//	isHit_m       (),
//	intersection_m(),
//	distance_m    ()
{
}


OctreeVisitorTracingSubpart::OctreeVisitorTracingSubpart
(
	const Scene*              pScene,
	const ObjectRenderableId& objectId,
	const Ray&                ray
)
 :	OctreeVisitorT<SubPart>(),
	pScene_m               ( pScene ),
//	pLastSubpart_m       (),
//	subpartUntransLower_m(),
//	subpartUntransUpper_m(),
//	//subpartTransBound_m[](),
//	subpartInvTransform_m()
	pObjectId_m   ( &objectId ),
	pRay_m        ( &ray ),
	isHit_m       ( false ),
	intersection_m(),
	distance_m    ()
{
}


OctreeVisitorTracingSubpart::~OctreeVisitorTracingSubpart()
{
}




/// commands ---------------------------------------------------------------------------------------
void OctreeVisitorTracingSubpart::visitRoot
(
	const OctreeCell* pRootCell,
	const OctreeData& octreeData
)
{
	/// assume intersection with octree root bound has succeeded

	/// move deeper into octree
	pRootCell->visitItems( octreeData, *this );
}


void OctreeVisitorTracingSubpart::visitBranch
(
	const OctreeCell* [],//subCells[8],
	const OctreeData& //octreeData
)
{
/*
	/// get which subcell is intersected


	do
	{
		/// move deeper into octree
		pSubcell->visitItems( subcellData, *this );

		/// if intersection has happened, climb back out of recursion
		if( isHit_m )
		{
			break;
		}

		/// move across to next subcell

	}
	while( not moved out of cell );
*/
}


void OctreeVisitorTracingSubpart::visitLeaf
(
	const Array<const ObjectInstance::SubPart*>& subparts,
	const OctreeData&                            octreeData
)
{
	const ObjectRenderableId& objectId = *pObjectId_m;
	const Ray&                ray      = *pRay_m;
	const Scene*              pScene   = pScene_m;

	bool isAnyHits = false;

	Intersection   nearestIntersection;
	float          nearestDistance = FLOAT_MAX;
	const SubPart* pNearestSubpart = 0;

	/// step through sub parts
	const SubPart** ppBegin = subparts.getMemory();
	const SubPart** ppEnd   = ppBegin + subparts.getLength();
	for( const SubPart** ppSubpart = ppBegin;  ppSubpart < ppEnd;  ++ppSubpart )
	{
		const SubPart& subpart = **ppSubpart;

		/// intersect
		Intersection intersectionTrial;
		float        distanceTrial;
		const bool isHit = subpart.getIntersection( ppSubpart - ppBegin, *pScene, objectId, ray, intersectionTrial, distanceTrial );

		/// check intersection is inside cell
		if( isHit && isInsideCell( octreeData.getBound(), intersectionTrial.position_m ) )
		{
			/// keep nearest hit data
			isAnyHits |= isHit;
			if( distanceTrial < nearestDistance )
			{
				nearestIntersection = intersectionTrial;
				nearestDistance     = distanceTrial;
				pNearestSubpart     = &subpart;
			}
		}
	}

	/// set intersection output params
	if( isAnyHits )
	{
		isHit_m        = isAnyHits;
		intersection_m = nearestIntersection;
		distance_m     = nearestDistance;

		/// outward transform intersection
		const Transform& transform = pNearestSubpart->getTransform();
		intersection_m.transformEq( transform );
	}
}




/// queries ----------------------------------------------------------------------------------------
bool OctreeVisitorTracingSubpart::isOverlappingCell
(
	const ObjectInstance::SubPart& subpart,
	const Vector3f&                lower,
	const Vector3f&                upper
) const
{
	updateOverlapCache( subpart );

	return isOverlapping( subpartTransBound_m, lower, upper );
}


dword OctreeVisitorTracingSubpart::getSubcellOverlaps
(
	const ObjectInstance::SubPart& subpart,
	const Vector3f&                lower,
	const Vector3f&                middle,
	const Vector3f&                upper
) const
{
	updateOverlapCache( subpart );

	/// return value of getSubcellOverlaps is 8 bits, each bit is a bool
	/// corresponding to a subcell, the high bit for subcell 7, the low bit for
	/// subcell 0.
	///
	/// subcell numbering:
	/// 		y z           110 111
	/// 		|/   010 011  100 101
	/// 		 -x  000 001

	dword overlapFlags = 0;

	/// test subpart against each subcell
	const Vector3f* cellCorners[] = { &lower, &middle, &upper };
	for( dword i = 8;  i-- > 0; )
	{
		/// make subcell lower and upper corners
		Vector3f subcellLower( cellCorners[ i & 1      ]->getX(),
		                       cellCorners[(i & 2) >> 1]->getY(),
		                       cellCorners[(i & 4) >> 2]->getZ() );
		Vector3f subcellUpper( cellCorners[((i & 1)     ) + 1]->getX(),
		                       cellCorners[((i & 2) >> 1) + 1]->getY(),
		                       cellCorners[((i & 4) >> 2) + 1]->getZ() );

		const bool is = isOverlapping( subpartTransBound_m, subcellLower, subcellUpper );
		overlapFlags |= dword(is) << i;
	}

	return overlapFlags;
}


bool OctreeVisitorTracingSubpart::getIsHit() const
{
	return isHit_m;
}


const Intersection& OctreeVisitorTracingSubpart::getIntersection() const
{
	return intersection_m;
}


float OctreeVisitorTracingSubpart::getDistance() const
{
	return distance_m;
}




/// implementation ---------------------------------------------------------------------------------
bool OctreeVisitorTracingSubpart::isInsideCell
(
	const OctreeBound& cell,
	const Vector3f&    position
)
{
	const Vector3f& cellLower = cell.getLowerCorner();
	const Vector3f& cellUpper = cell.getUpperCorner();

	const bool isInside = (position.getX() >= cellLower.getX()) &
	                      (position.getX() <  cellUpper.getX()) &
	                      (position.getY() >= cellLower.getY()) &
	                      (position.getY() <  cellUpper.getY()) &
	                      (position.getZ() >= cellLower.getZ()) &
	                      (position.getZ() <  cellUpper.getZ());
	return isInside;
}


void OctreeVisitorTracingSubpart::updateOverlapCache
(
	const ObjectInstance::SubPart& subpart
) const
{
	/// recalculate cache for different subpart
	if( &subpart != pLastSubpart_m )
	{
		const ObjectRenderable* const pObject = pScene_m->getObject( subpart.getObjectIndex() );
		const Bound&      subpartBound     = pObject->getBound();
		const Transplace& subpartTransform = subpart.getTransform();


		subpartBound.getMinMax( subpartUntransLower_m, subpartUntransUpper_m );

		subpartBound.getCorners( subpartTransBound_m );
		for( int i = 8;  i-- > 0; )
		{
			subpartTransform.transform( subpartTransBound_m[i] );
		}

		subpartInvTransform_m = subpartTransform;
		subpartInvTransform_m.invert();

		pLastSubpart_m = &subpart;
	}
}


bool OctreeVisitorTracingSubpart::isOverlapping
(
	const Vector3f  subpartTransBound[8],
	const Vector3f& cellLower,
	const Vector3f& cellUpper
) const
{
	bool isAllOutside = isAllOutsideAABB( subpartTransBound, cellLower, cellUpper );

	/// decide indeterminate result by swapping boxs and re-testing in the alternate boxs coordinate space
	if( !isAllOutside )
	{
		const Vector3f* cellEnds[] = { &cellLower, &cellUpper };

		/// make bound corner points from lower and upper corners, and transforming
		Vector3f cellInvTransBound[8];
		for( dword i = 8;  i-- > 0; )
		{
			cellInvTransBound[i].setXYZ( cellEnds[ i & 1      ]->getX(),
			                             cellEnds[(i & 2) >> 1]->getY(),
			                             cellEnds[(i & 4) >> 2]->getZ() );
			subpartInvTransform_m.transform( cellInvTransBound[i] );
		}

		/// re-test
		isAllOutside = isAllOutsideAABB( cellInvTransBound, subpartUntransLower_m, subpartUntransUpper_m );
	}

	return !isAllOutside;
}


bool OctreeVisitorTracingSubpart::isAllOutsideAABB
(
	const Vector3f  points[8],
	const Vector3f& aabbLower,
	const Vector3f& aabbUpper
)
{
	dword allOneSideFlags[6] = { dword(true), dword(true), dword(true),
	                             dword(true), dword(true), dword(true) };

	/// find if all points are outside any AABB planes
	/// and therefore not overlapping the AABB
	for( dword i = 8;  i-- > 0; )
	{
		const Vector3f& point = points[i];

		allOneSideFlags[0] &= dword(point.getX() <  aabbLower.getX());
		allOneSideFlags[1] &= dword(point.getY() <  aabbLower.getY());
		allOneSideFlags[2] &= dword(point.getZ() <  aabbLower.getZ());
		allOneSideFlags[3] &= dword(point.getX() >= aabbUpper.getX());
		allOneSideFlags[4] &= dword(point.getY() >= aabbUpper.getY());
		allOneSideFlags[5] &= dword(point.getZ() >= aabbUpper.getZ());
	}

	return (allOneSideFlags[0] | allOneSideFlags[1] | allOneSideFlags[2] |
	        allOneSideFlags[3] | allOneSideFlags[4] | allOneSideFlags[5]) != 0;


//	dword allOneSideFlags = 0xFFFFFFFF;
//
//	for( dword i = 8;  i-- > 0; )
//	{
//		const Vector3f& point = points[i];
//
//		const dword lowerFlags =  (point.getX() <  aabbLower.getX())       |
//		                         ((point.getY() <  aabbLower.getY()) << 1) |
//		                         ((point.getZ() <  aabbLower.getZ()) << 2)
//		const dword upperFlags =  (point.getX() >= aabbUpper.getX())       |
//		                         ((point.getY() >= aabbUpper.getY()) << 1) |
//		                         ((point.getZ() >= aabbUpper.getZ()) << 2)
//		allOneSideFlags &= (lowerFlags | (upperFlags << 3));
//	}
//
//	return allOneSideFlags != 0;
}
