/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef OctreeVisitorTracingSubpart_h
#define OctreeVisitorTracingSubpart_h


#include "Primitives.h"
#include "ObjectInstance.h"
#include "Intersection.h"
#include "Octree.h"
#include "Vector3f.h"
#include "Transplace.h"

#include "hxa7241_graphics.h"




#include "perceptuum2_modelling.h"
namespace perceptuum2 { namespace modelling {
	using namespace hxa7241::primitives;
	using perceptuum2::modelling::ObjectInstance::SubPart;
	using hxa7241::graphics::OctreeVisitorT;
	using hxa7241::graphics::OctreeVisitor;
	using hxa7241::graphics::OctreeData;
	using hxa7241::graphics::OctreeBound;
	using hxa7241::graphics::OctreeCell;
	using hxa7241::graphics::Vector3f;
	using hxa7241::graphics::Transplace;
	using hxa7241::utility::Array;


/**
 * an octree visitor for tracing object instance sub parts.<br/><br/>
 *
 * @implementation
 * maybe improved by State pattern, to separate overlap operations from
 * visit operations.
 */
class OctreeVisitorTracingSubpart
	: public OctreeVisitorT<SubPart>
{
/// standard object services -----------------------------------------------------------------------
public:
	         /**
	          * for inserting or removing.
	          */
	         OctreeVisitorTracingSubpart( const Scene* pScene );
	         /**
	          * for visiting.
	          */
	         OctreeVisitorTracingSubpart( const Scene*              pScene,
	                                      const ObjectRenderableId& objectId,
	                                      const Ray&                ray );

	virtual ~OctreeVisitorTracingSubpart();
private:
	         OctreeVisitorTracingSubpart( const OctreeVisitorTracingSubpart& );
	OctreeVisitorTracingSubpart& operator=( const OctreeVisitorTracingSubpart& );


/// octree visitor overrides
/// commands ---------------------------------------------------------------------------------------
protected:
	virtual void  visitRoot  ( const OctreeCell* pRootCell,
	                           const OctreeData& octreeData );
	virtual void  visitBranch( const OctreeCell* subCells[8],
	                           const OctreeData& octreeData );
	virtual void  visitLeaf  ( const Array<const SubPart*>& items,
	                           const OctreeData&            octreeData );


/// queries ----------------------------------------------------------------------------------------
	virtual bool  isOverlappingCell ( const SubPart&  item,
	                                  const Vector3f& lowerCorner,
	                                  const Vector3f& upperCorner )                           const;
	virtual dword getSubcellOverlaps( const SubPart&  item,
	                                  const Vector3f& lower,
	                                  const Vector3f& middle,
	                                  const Vector3f& upper )                                 const;

/// specialised interface
public:
	virtual bool                getIsHit()                                                    const;
	virtual const Intersection& getIntersection()                                             const;
	virtual float               getDistance()                                                 const;


/// implementation ---------------------------------------------------------------------------------
protected:
	static  bool  isInsideCell( const OctreeBound& cell,
	                            const Vector3f&    position );

	virtual void  updateOverlapCache( const SubPart& )                                        const;

	virtual bool  isOverlapping( const Vector3f  subpartTransBound[8],
	                             const Vector3f& cellLower,
	                             const Vector3f& cellUpper )                                  const;
	/** @return true means outside, false means undetermined*/
	static  bool  isAllOutsideAABB( const Vector3f  points[8],
	                                const Vector3f& aabbLower,
	                                const Vector3f& aabbUpper );


/// fields -----------------------------------------------------------------------------------------
private:
	const Scene*const pScene_m;

	/// overlap state ----------------------------
	/// calculation cache
	mutable const SubPart* pLastSubpart_m;
	mutable Vector3f       subpartUntransLower_m;
	mutable Vector3f       subpartUntransUpper_m;
	mutable Vector3f       subpartTransBound_m[8];
	mutable Transplace     subpartInvTransform_m;


	/// visit state ------------------------------
	/// initial ray data
	const ObjectRenderableId* pObjectId_m;
	const Ray*                pRay_m;

	/// nearest intersection data
	bool         isHit_m;
	Intersection intersection_m;
	float        distance_m;
};


}}




#endif//OctreeVisitorTracingSubpart_h
