/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Projector.h"
#include "SpacetimeIndex.h"
#include "Projection.h"
#include "Illumination.h"
#include "Light.h"
#include "ObjectRenderableId.h"
#include "ObjectRenderable.h"
#include "Scene.h"
#include "Ray.h"
#include "ImageAccumulation.h"
#include "CieXyz.h"
#include "Transclip.h"

#include "RasterizerProjective.h"   /// own header is included last


using namespace perceptuum2::transport::projection;




/// standard object services -----------------------------------------------------------------------
RasterizerProjective::RasterizerProjective
(
	const Projection& projection,
	const dword       imageWidth,
	const dword       imageHeight,
	const dword       ,//eyePathLengthLimit,
	const bool        isLogging
)
 :	camera_m( projection, imageWidth, imageHeight ),
	objectBuffer_m(),
	meshBuffer_m(),
	rasterVertexBuffer_m(),
	rasterTriangleBuffer_m(),
	isLogging_m( isLogging )
{
}


RasterizerProjective::~RasterizerProjective()
{
}




/// commands ---------------------------------------------------------------------------------------
void RasterizerProjective::rasterizeFrame
(
	const SamplingControl& samplingControl,
	const SpacetimeIndex&  sceneIndex,
	const Illumination&    ,//illumination,
	ImageAccumulation&     //accumulation
)
{
	/// prepare projector
	Projector projector( camera_m, samplingControl );


	/// get objects from scene
	{
		const ObjectRenderableId rootObject( sceneIndex.getScene().getRootObject() );

		/// prepare object storage
		objectBuffer_m.resize( 0 );
		objectBuffer_m.reserve( rootObject->getSubPartCount() );

		/// fill objects collection
		const hxa7241::graphics::Transclip transclip( projector.getWorldToClip() );
		rootObject->getObjectsForProjection( transclip, ClipFlags::ALL(), objectBuffer_m );
	}


	/// step through objects, rasterizing each
	for( dword i = 0, end = objectBuffer_m.size();  i < end;  ++i )
	{
		const ProjectableObjectRef& object = objectBuffer_m[i];
		if( object.getObject() != 0 )
		{
			/// get object mesh
			const Mesh* pMesh = 0;
			object.getObject()->getMesh( meshBuffer_m, &pMesh );

			if( pMesh != 0 )
			{
				/// set transforms
				projector.setModelToClip( object.getTransform().getMatrix() );

				/// transform model vertexs into world and screen space
				transformVertexs( projector, object.getClipFlags(), *pMesh, rasterVertexBuffer_m );

				/// select visible triangles
				selectTriangles( projector, pMesh->getTriangles(), rasterTriangleBuffer_m, rasterVertexBuffer_m );

				/// rasterize mesh
//				rasterizeMesh( rasterVertexBuffer_m, rasterTriangleBuffer_m,
//				               *(object.getObject()), eyePosition,
//				               sceneIndex, illumination, accumulation );
			}
		}
	}
}




/// implementation ---------------------------------------------------------------------------------
void RasterizerProjective::transformVertexs
(
	const Projector&             projector,
	const ClipFlags              objectClipflags,
	const Mesh&                  mesh,
	Vector<RasterizationVertex>& rasterVertexs
) /// static
{
	/// get vertex context data
	const vector<Vertex>& modelVertexs = mesh.getVertexs();
	const Vector3f&       meshPosition = mesh.getPosition();
	const Vector3f&       meshScale    = mesh.getScale();

	/// prepare buffer (allowing for clipping to add vertexs later (+10%))
	{
		dword vertexsSize = dword(modelVertexs.size());
		vertexsSize += vertexsSize / 10;
		rasterVertexs.setUsedSize( vertexsSize );
	}

	/// prepare to step through model vertexs
	const Vertex*        pModelVertex  = &(modelVertexs.front());
	const Vertex*        pEnd          = pModelVertex + modelVertexs.size();
	RasterizationVertex* pRasterVertex = rasterVertexs.getFront();

//	if( objectClipflags.isUnclipped() )
//	{
//		/// accelerate unclipped case
//		for( ;  pModelVertex < pEnd;  ++pModelVertex, ++pRasterVertex )
//		{
//			/// convert vertex
//			pRasterVertex->set( *pModelVertex, meshPosition, meshScale, projector );
//		}
//	}
//	else
	{
		/// general case
		for( ;  pModelVertex < pEnd;  ++pModelVertex, ++pRasterVertex )
		{
			/// convert vertex
			pRasterVertex->set( *pModelVertex, meshPosition, meshScale, objectClipflags, projector );
		}
	}
}


void RasterizerProjective::selectTriangles
(
	const Projector&                 ,//projector,
	const vector<TriangleImplicit>&  ,//triangles,
	Vector<RasterizationTriangleIm>& ,//rasterTriangles,
	Vector<RasterizationVertex>&     //rasterVertexs
) /// static
{
/*	/// prepare buffer (allowing for clipping to add triangles (+10%))
	{
		dword trianglesSize = dword(triangles().size());
		trianglesSize += trianglesSize / 10;
		rasterTriangles.setUsedSize( trianglesSize );
	}

	/// step through triangles
	const TriangleImplicit*  pTriangle       = &(triangles.front());
	const TriangleImplicit*  pEnd            = pTriangle + triangles.size();
	RasterizationTriangleIm* pRasterTriangle = rasterTriangles.getFront();
	for( dword pTriangleId = 0;  pTriangle < pEnd;  ++pTriangle )
	{
		/// get triangle vertexs
		const dword                triVertIndexs[3] = { pTriangle->getVertexIndex0(),    pTriangle->getVertexIndex1(),    pTriangle->getVertexIndex2() };
		const RasterizationVertex& triVertexs[3]    = { rasterVertexs[triVertIndexs[0]], rasterVertexs[triVertIndexs[1]], rasterVertexs[triVertIndexs[2]] };

		const ClipFlags vertexClipFlagsOr( triVertexs[0].getClipFlags() | triVertexs[1].getClipFlags() | triVertexs[2].getClipFlags() );

		/// check for unclipped
		if( vertexClipFlagsOr.isUnclipped() )
		{
			/// check not backfacing
			if( RasterizationTriangleIm::isFrontFacing( triVertexs ) )
			{
				/// add triangle
				pRasterTriangle->set( *pTriangle, pTriangleId )
				++pRasterTriangle;
				++pTriangleId;
			}
		}
		/// check for not entirely clipped
		else if( (triVertexs[0].getClipFlags() & triVertexs[1].getClipFlags() & triVertexs[2].getClipFlags()).isUnclipped() )
		{
			/// check not backfacing, but only if not clipped by near z plane
			/// (havent decided whether the backfacing test works for triangles straddling the perspective z cusp...)
			if( vertexClipFlagsOr.isClipped( ClipFlags::ZLower ) ||
			    RasterizationTriangleIm::isFrontFacing( triVertexs ) )
			{
				/// make clip coords of vertexs
				ClipCoords triVertexsClipCoords[3];
				for( int i = 3;  i-- > 0; )
				{
					/// transform vertex position to clip space
					Vector4f triVertexsClipSpace;
					projector.transformWorldToClip( triVertexs[i].getWorldPosition(), triVertexsClipSpace );

					triVertexsClipCoords[i].setForPoint( triVertexsClipSpace );
				}

				/// clip triangle
					/// step through dimensions
						/// step through vertexs
							/// interpolate vertex positions and texture coords
							/// add to vertexs list
					/// triangulate polygon
						/// add triangles and vertexs to buffers
							/// check for buffer overflows, tri and vert
			}
		}
	}*/
}


void RasterizerProjective::rasterizeMesh
(
	const Vector<RasterizationVertex>&     ,//vertexs,
	const Vector<RasterizationTriangleIm>& ,//triangles,
	const ObjectRenderable&                ,//object,
	const Vector3f&                        ,//eyePosition,
	const SpacetimeIndex&                  ,//sceneIndex,
	const Illumination&                    ,//illumination,
	ImageAccumulation&                     //accumulation
)
{
}








/*
	class Projector
	{
		Projector( Camera, SamplingControl )
		setModelToClip( const Transclip& )

		getWorldToClip( Transclip& )
		transformModelToScreen( ClipFlags, const Vector3f&, Vector3f&, ClipFlags& )
		transformModelToWorld( const Vector3f&, Vector3f& )
		transformWorldToClip( const Vector3f&, Vector4f& )

		/// screen space is fp lefttop zero and pixel scale
		/// clip space is blinn preperspective homogeneous 0-1 cube
		/// world space is root scene object

		Matrix4f worldToClip_m;
		Matrix4f modelToClip_m;
		Matrix4f modelToWorld_m;
		         clipToScreen_m;
	}

	class RasterizationVertex
	{
		void set( const Vertex&, const Vector3f& meshPosition, const Vector3f& meshScale, ClipFlags, const Projector& );

		const Vector3f& getScreenPosition() const;
		const Vector3f& getWorldPosition()  const;
		ClipFlags       getClipFlags()      const;

		Vector3f  worldPosition_m;
		Vector3f  screenPosition_m;
		// float     screenW_m; ?
		ClipFlags clipFlags_m;
	}

	class RasterizationTriangleIm
		: public TriangleImplicit
	{
		void set( const TriangleImplicit&, dword id );

		bool  isFrontFacing( const RasterizationVertex& verts[3] ) const;
		{
			const Vector3& vertScreenPoss[3] = { triVertexs[0].getScreenPosition(), triVertexs[1].getScreenPosition(), triVertexs[2].getScreenPosition() };
			return ( ( (vertScreenPoss[2].getX() - vertScreenPoss[0].getX()) * (vertScreenPoss[1].getY() - vertScreenPoss[0].getY()) ) <
			         ( (vertScreenPoss[2].getY() - vertScreenPoss[0].getY()) * (vertScreenPoss[1].getX() - vertScreenPoss[0].getX()) ) )
		}
		dword getId() const;

		dword id_m;
	}

	class RasterizationCorner
	{
		RasterizationCorner( const RasterizationVertex&, const Vector2<word>& );
		void illuminate( const SpacetimeIndex&, const ObjectRenderableId& );

		Vector3f  worldPosition_m;
		Vector3f  screenPosition_m;
		float     screenW_m;
		ClipFlags clipFlags_m;

		Vector2<word> textureCoords_m;

		Light exitLightOverReflectivity_m;
	}

	class RasterizationTriangleEx
	{
		RasterizationTriangleEx( const RasterizationCorner&, const RasterizationCorner&, const RasterizationCorner& )
		void illuminate( const SpacetimeIndex&, const ObjectRenderableId& );

		void draw( const ObjectRenderbleId&, const Vector3f& eyePosition, const SpacetimeIndex&, const Illumination&, ImageAccumulation& ) const;
		void fill( const ObjectRenderbleId&, const SpacetimeIndex&, const Illumination&, ImageAccumulation& ) const;
		void scanConvert( ImageAccumulation& ) const;
		void traceInto( ImageAccumulation& ) const;

		const RasterizationCorner* pCorner0_m;
		const RasterizationCorner* pCorner1_m;
		const RasterizationCorner* pCorner2_m;
	}


void RasterizerProjective::rasterizeFrame
(
	const SamplingControl& samplingControl,
	const SpacetimeIndex&  sceneIndex,
	const Illumination&    illumination,
	ImageAccumulation&     accumulation
)
{
	/// prepare projector
	Projector projector( camera, samplingControl )


	/// get objects from scene
	{
		const ObjectRenderableId rootObject( sceneIndex.getScene().getRootObject() );

		/// prepare object storage
		objectBuffer_m.resize( 0 );
		objectBuffer_m.reserve( rootObject->getSubPartCount() );

		/// fill objects collection
		Transclip transclip;
		projector.getWorldToClip( transclip )
		rootObject->getObjectsForProjection( transclip, ClipFlags::ALL(), objectBuffer_m );
	}


	/// step through objects, rasterizing each
	for( dword i = 0, end = objectBuffer_m.size();  i < end;  ++i )
	{
		const ProjectableObjectRef& object = objectBuffer_m[i];
		if( object.pObject_m != 0 )
		{
			/// get object mesh
			const Mesh* pMesh = 0;
			object.pObject_m->getMesh( meshBuffer_m, &pMesh );

			/// set transforms
			projector.setModelToClip( object.transform_m )

			if( pMesh != 0 )
			{
				/// transform model vertexs into world and screen space
				transformVertexs( projector, object.clipFlags_m, *pMesh, rasterVertexBuffer_m );

				/// select visible triangles
				selectTriangles( projector, pMesh->getTriangles(), rasterTriangleBuffer_m, rasterVertexBuffer_m );

				/// rasterize mesh
				rasterizeMesh( rasterVertexBuffer_m, rasterTriangleBuffer_m,
				               *(object.pObject_m), eyePosition,
				               sceneIndex, illumination, accumulation );
			}
		}
	}
}




void RasterizerProjective::rasterizeMesh
(
	const Vector<RasterizationVertex>&     vertexs,
	const Vector<RasterizationTriangleIm>& triangles,
	const ObjectRenderable&                object,
	const Vector3f&                        eyePosition,
	const SpacetimeIndex&                  sceneIndex,
	const Illumination&                    illumination,
	ImageAccumulation&                     accumulation
)
{
	/// step through triangles
	const RasterizationTriangleIm* pTriangle = &(triangles.front());
	const RasterizationTriangleIm* pEnd      = pTriangle + triangles.size();
	for( dword t = 0;  pTriangle < pEnd;  ++pTriangle, ++t )
	{
		/// get texture coords
		Vector2<word> txy[3];
		pTriangle->getTextureCoords( txy );

		/// make corners
		const RasterizationCorner corner0( vertexs[pTriangle->getVertexIndex0()], txy[0] );
		const RasterizationCorner corner1( vertexs[pTriangle->getVertexIndex1()], txy[1] );
		const RasterizationCorner corner2( vertexs[pTriangle->getVertexIndex2()], txy[2] );

		/// make a rasterization triangle
		const RasterizationTriangleEx rasterTriangle( corner0, corner1, corner2 );

		const ObjectRenderableId objectId( &object, t );

		/// illuminate triangle
		rasterTriangle.illuminate( sceneIndex, objectId, eyePosition );

		/// rasterize triangle
		rasterTriangle.draw( objectId, eyePosition, sceneIndex, illumination, accumulation );
	}
}



void RasterizationTriangleEx::illuminate
(
	const SpacetimeIndex&     sceneIndex,
	const ObjectRenderableId& objectId,
	const Vector3f&           eyePosition
)
{
	pCorner0_m.illuminate( sceneIndex, objectId, eyePosition );
	pCorner1_m.illuminate( sceneIndex, objectId, eyePosition );
	pCorner2_m.illuminate( sceneIndex, objectId, eyePosition );
}


void RasterizationCorner::illuminate
(
	const SpacetimeIndex&     sceneIndex,
	const ObjectRenderableId& objectId,
	const Vector3f&           eyePosition
)
{
	/// make an eye path node at the corner
	EyePathNode eyePathNode;
	{
/// is this normal the correct one for an intersection? no, it should be 'structural' - not interpolated or bumped
/// ?
		const Intersection       intersection( worldPosition_m, worldNormal_m, textureCoords_m, objectId );
		const ObjectRenderableId outsideObject( sceneIndex.getScene().getDefaultObject() );

		eyePathNode.setSurfacePoint( intersection, outsideObject );
	}

	/// get light toward eye
	Light eyeLight;
	{
		const UnitVector3f eyeDirection( worldPosition_m - eyePosition );

		illumination.getIllumination( sceneIndex, eyePathNode, eyeDirection, eyeLight );
	}

	/// divide light by reflectivity
	Light reflectivity;
	eyePathNode.getReflectivity( reflectivity )
	eyeLight /= reflectivity;   /// modify Light::op/= to avoid div by zero


	/// set
	exitLightOverReflectivity_m = eyeLight;
}


void RasterizationTriangleEx::draw
(
	const ObjectRenderableId& objectId,
	const Vector3f&           eyePosition,
	const SpacetimeIndex&     sceneIndex,
	const Illumination&       illumination,
	ImageAccumulation&        accumulation
) const
{
	/// if tri is, hmmm, > 16 pixel area maybe, or has a side > 5 pixels might be better
		/// calc three half edge verts (maybe with stochastic imprecision...
			/// no, must do exactly the same calculation on the same numbers, or adjacent tris will overlap or gap.
			/// yes, but the pos sent to illum could be nudged stochastically)
		/// make eyepathnode and call illuminate for each half edge vert, by inv-transforming pos
		/// step through sub-tris calling subdivide on each, if its not clipped entirely
	/// else
		/// fill triangle
}


void RasterizationTriangleEx::fill
(
	const ObjectRenderableId& objectId,
	const Illumination&       illumination,
	ImageAccumulation&        accumulation
) const
{
	/// make eye path node for each corner
	/// if all corners below shinyness threshold, scan convert
		/// starting with light out / refectivity, texture coords, at each vert
		/// interpolate xyzw, uv, light as cie
			/// emit has to be handled specially
	/// else, trace into
		/// interpolate xyz, norm, uv
		/// make eyepathnode each pixel
		/// trace
}


void RasterizationTriangleEx::traceInto
(
	ImageAccumulation& accumulation
) const
{
}


void RasterizationTriangleEx::scanConvert
(
	Sheet<float>&      zBuffer,
	ImageAccumulation& accumulation
) const
{


	/// at each corner:  x,y,z,p, u,v, light
	///


	const float xFractional = ;
	const float yFractional = ;

	const dword imageWidth = accumulation.getWidth();

	const dword  yStarts[2];
	const dword  yStops[2];
	const dword  xLeftStarts[2];
	const dword  xRightStarts[2];
	const float  pLeftStarts[2];
	const float  zLeftStarts[2];
	const CieXyz lightLeftStarts[2];
	const float  pIncStarts[2];
	const float  zIncStarts[2];
	const CieXyz lightIncStarts[2];

	const dword  xLeftIncStarts[2];
	const dword  xRightIncStarts[2];
	const float  pLeftIncStarts[2];
	const float  zLeftIncStarts[2];
	const CieXyz lightLeftIncStarts[2];
	const float  pIncIncStarts[2];
	const float  zIncIncStarts[2];
	const CieXyz lightIncIncStarts[2];

	MaterialIterator materialStarts[2];


	/// step through two trapezoids
	for( dword t = 0;  t <= 1;  ++t )
	{
		/// check for empty trapezoid
		if( yStarts[t] < yStops[t] )
		{
			/// prepare trapezoid interpolants
			dword y          = yStarts[t];
			const dword yEnd = yStops[t];

			dword xLeft  = xLeftStarts[t];
			dword xRight = xRightStarts[t];

			float* pZBufferLeft = &( zBuffer.getElement( 0, yStarts[t] ) );
			float  pLeft = pLeftStarts[t];
			float  zLeft = zLeftStarts[t];
			CieXyz lightLeft( lightLeftStarts[t] );
			MaterialIterator& material = materialStarts[t];

			float  pInc = pIncStarts[t];
			float  zInc = zIncStarts[t];
			CieXyz lightInc( lightIncStarts[t] );

			/// prepare trapezoid increments
			const dword xLeftInc  = xLeftIncStarts[t];
			const dword xRightInc = xRightIncStarts[t];

			const float  pLeftInc = pLeftIncStarts[t];
			const float  zLeftInc = zLeftIncStarts[t];
			const CieXyz lightLeftInc( lightLeftIncStarts[t] );

			const float  pIncInc = pIncIncStarts[t];
			const float  zIncInc = zIncIncStarts[t];
			const CieXyz lightIncInc( lightIncIncStarts[t] );

			/// step through trapezoid of spans
			while( true )
			{
				/// check for empty span
				if( (xLeft >> 16) < (xRight >> 16) )
				{
					/// prepare pixel span interpolants
					dword x          = xLeft  >> 16;
					const dword xEnd = xRight >> 16;

					float* pZBuffer = pZBufferLeft + (x >> 16);
					float  p        = pLeft;
					float  z        = zLeft;
					CieXyz light( lightLeft );

					/// step along span of pixels
					while( true )
					{
						const float oneOverP = 1.0f / p;

						/// make pixel color, from texel reflectivity scaled by interpolated light
						CieXyz color( light );
						{
							CieXyz reflectivity;
							material.getReflectivity( oneOverP, reflectivity );

							(color *= oneOverP) *= reflectivity;
						}

						/// do z buffer checked write
						{
							const float zz = z * oneOverP;
							if( zz < *pZBuffer )
							{
								/// accumulate pixel color
								accumulation.addSample( x, y, xFractional, yFractional, color, FLOAT_ALMOST_ONE );

								/// set z buffer
								*pZBuffer = zz;
							}
						}

						/// test loop here to avoid increments on last iteration
						++x;
						if( x >= xEnd )
						{
							break;
						}

						/// increment interpolants
						++pZBuffer;
						p     += pInc;
						z     += zInc;
						light += lightInc;
						material.incrementX();
					}
				}

				/// test loop here to avoid increments on last iteration
				y += imageWidth;
				if( y >= yEnd )
				{
					break;
				}

				/// increment edges
				xLeft  += xLeftInc;
				xRight += xRightInc;

				pZBufferLeft += imageWidth;
				pLeft        += pLeftInc;
				zLeft        += zLeftInc;
				lightLeft    += lightLeftInc;
				material.incrementY();

				/// increment span increments
				pInc     += pIncInc;
				zInc     += zIncInc;
				lightInc += lightIncInc;
			}
		}
	}
}
*/








/*

	const ObjectRenderableId outsideObject( sceneIndex.getScene().getDefaultObject() );
	eyePathNode.setSurfacePoint( intersection, outsideObject );

	UnitVector3f directionFromEye( initialRay.getDirection() );
	UnitVector3f directionToEye( -initialRay.getDirection() );

	Light light;

	/// illuminate
	if( eyePathNode.getSpecularness() <= samplingControl.getSamplingPoint(samplingControl.EYE_PATH) )
	{
		illumination.getIllumination( sceneIndex, eyePathNode, directionToEye, light );
	}
	/// trace into
	else
	{
		/// prepare eye path node
		{
			using hxa7241::utility::Float01;
			using hxa7241::graphics::Vector2;
			const Vector2<Float01> rands( samplingControl.getSamplingPoint(samplingControl.EYE_INTERACTION_X), samplingControl.getSamplingPoint(samplingControl.EYE_INTERACTION_Y) );
			const Float01          roulette( samplingControl.getSamplingPoint(samplingControl.EYE_INTERACTION_T) );

			eyePathNode.addInwardPathStep( directionFromEye, rands, roulette );
		}

		/// trace from eye path node
		eyePath.traceFromFirstHit( samplingControl, sceneIndex, illumination, eyePathNode );

		/// get path value at eye end of path
		eyePath.getEyeEndLight( light );

		///
		eyePathNode.propagateLight( light, directionToEye, light );
	}

*/
